@@ -18,43 +18,27 @@
 
 # Must happen before any importing of tensorflow to curtail mem usage
 from merlin.models.loader.tf_utils import configure_tensorflow
-from merlin.models.tf.utils.tf_utils import TensorInitializer
+from merlin.models.tf.core.index import IndexBlock, TopKIndexBlock
+from merlin.models.tf.core.tabular import AsTabular, Filter, TabularBlock
+from merlin.models.tf.core.transformations import (
+    AsDenseFeatures,
+    AsRaggedFeatures,
+    AsSparseFeatures,
+    CategoricalOneHot,
+    ExpandDims,
+    HashedCross,
+    LabelToOneHot,
+)
 
 configure_tensorflow()
 
 from tensorflow.keras.layers import Dense, Layer
-from tensorflow.python.keras.losses import Loss
-from tensorflow.python.keras.metrics import Metric
-from tensorflow.python.keras.optimizer_v2.optimizer_v2 import OptimizerV2
+from tensorflow.keras.losses import Loss
+from tensorflow.keras.metrics import Metric
+from tensorflow.keras.optimizers import Optimizer
 from tensorflow.python.training.tracking.data_structures import ListWrapper, _DictWrapper
 
 from merlin.models.loader.tf_utils import configure_tensorflow
-from merlin.models.tf.blocks.core.aggregation import (
-    ConcatFeatures,
-    ElementwiseSum,
-    ElementwiseSumItemMulti,
-    StackFeatures,
-)
-from merlin.models.tf.blocks.core.base import (
-    Block,
-    EmbeddingWithMetadata,
-    ModelContext,
-    NoOp,
-    right_shift_layer,
-)
-from merlin.models.tf.blocks.core.combinators import ParallelBlock, ResidualBlock, SequentialBlock
-from merlin.models.tf.blocks.core.context import FeatureContext
-from merlin.models.tf.blocks.core.index import IndexBlock, TopKIndexBlock
-from merlin.models.tf.blocks.core.masking import CausalLanguageModeling, MaskedLanguageModeling
-from merlin.models.tf.blocks.core.tabular import AsTabular, Filter, TabularBlock
-from merlin.models.tf.blocks.core.transformations import (
-    AsDenseFeatures,
-    AsSparseFeatures,
-    CategoricalOneHot,
-    ExpandDims,
-    LabelToOneHot,
-    StochasticSwapNoise,
-)
 from merlin.models.tf.blocks.cross import CrossBlock
 from merlin.models.tf.blocks.dlrm import DLRMBlock
 from merlin.models.tf.blocks.experts import CGCBlock, MMOEBlock, MMOEGate
@@ -67,13 +51,24 @@ from merlin.models.tf.blocks.retrieval.matrix_factorization import (
 )
 from merlin.models.tf.blocks.retrieval.two_tower import TwoTowerBlock
 from merlin.models.tf.blocks.sampling.base import ItemSampler
-from merlin.models.tf.blocks.sampling.cross_batch import (
-    CachedCrossBatchSampler,
-    CachedUniformSampler,
-    PopularityBasedSampler,
-)
+from merlin.models.tf.blocks.sampling.cross_batch import PopularityBasedSampler
 from merlin.models.tf.blocks.sampling.in_batch import InBatchSampler
 from merlin.models.tf.blocks.sampling.queue import FIFOQueue
+from merlin.models.tf.core.aggregation import (
+    ConcatFeatures,
+    ElementwiseSum,
+    ElementwiseSumItemMulti,
+    StackFeatures,
+)
+from merlin.models.tf.core.base import (
+    Block,
+    EmbeddingWithMetadata,
+    ModelContext,
+    NoOp,
+    right_shift_layer,
+)
+from merlin.models.tf.core.combinators import Cond, ParallelBlock, ResidualBlock, SequentialBlock
+from merlin.models.tf.data_augmentation.noise import StochasticSwapNoise
 from merlin.models.tf.dataset import sample_batch
 from merlin.models.tf.inputs.base import InputBlock
 from merlin.models.tf.inputs.continuous import ContinuousFeatures
@@ -81,21 +76,22 @@ from merlin.models.tf.inputs.embedding import (
     ContinuousEmbedding,
     EmbeddingFeatures,
     EmbeddingOptions,
+    EmbeddingTable,
     FeatureConfig,
     SequenceEmbeddingFeatures,
     TableConfig,
 )
 from merlin.models.tf.losses import LossType
-from merlin.models.tf.metrics.ranking import (
+from merlin.models.tf.metrics.topk import (
     AvgPrecisionAt,
     MRRAt,
     NDCGAt,
     PrecisionAt,
     RecallAt,
-    ranking_metrics,
+    TopKMetricsAggregator,
 )
 from merlin.models.tf.models import benchmark
-from merlin.models.tf.models.base import Model, RetrievalModel
+from merlin.models.tf.models.base import BaseModel, Model, RetrievalModel
 from merlin.models.tf.models.ranking import DCNModel, DeepFMModel, DLRMModel
 from merlin.models.tf.models.retrieval import (
     MatrixFactorizationModel,
@@ -112,6 +108,7 @@ from merlin.models.tf.prediction_tasks.next_item import NextItemPredictionTask
 from merlin.models.tf.prediction_tasks.regression import RegressionTask
 from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask
 from merlin.models.tf.utils import repr_utils
+from merlin.models.tf.utils.tf_utils import TensorInitializer
 
 ListWrapper.__repr__ = repr_utils.list_wrapper_repr
 _DictWrapper.__repr__ = repr_utils.dict_wrapper_repr
@@ -121,10 +118,11 @@ Layer.__rrshift__ = right_shift_layer
 Layer.__repr__ = repr_utils.layer_repr
 Loss.__repr__ = repr_utils.layer_repr_no_children
 Metric.__repr__ = repr_utils.layer_repr_no_children
-OptimizerV2.__repr__ = repr_utils.layer_repr_no_children
+Optimizer.__repr__ = repr_utils.layer_repr_no_children
 
 __all__ = [
     "Block",
+    "Cond",
     "ModelContext",
     "SequentialBlock",
     "ResidualBlock",
@@ -132,15 +130,12 @@ __all__ = [
     "CrossBlock",
     "DLRMBlock",
     "MLPBlock",
-    "CausalLanguageModeling",
-    "MaskedLanguageModeling",
     "ContinuousEmbedding",
     "MMOEGate",
     "MMOEBlock",
     "CGCBlock",
     "TopKIndexBlock",
     "IndexBlock",
-    "FeatureContext",
     "DenseResidualBlock",
     "TabularBlock",
     "ContinuousFeatures",
@@ -154,8 +149,10 @@ __all__ = [
     "MatrixFactorizationBlock",
     "QueryItemIdsEmbeddingsBlock",
     "AsDenseFeatures",
+    "AsRaggedFeatures",
     "AsSparseFeatures",
     "CategoricalOneHot",
+    "HashedCross",
     "ElementwiseSum",
     "ElementwiseSumItemMulti",
     "AsTabular",
@@ -178,7 +175,8 @@ __all__ = [
     "MRRAt",
     "AvgPrecisionAt",
     "RecallAt",
-    "ranking_metrics",
+    "BaseModel",
+    "TopKMetricsAggregator",
     "Model",
     "RetrievalModel",
     "InputBlock",
@@ -189,8 +187,6 @@ __all__ = [
     "ItemSampler",
     "EmbeddingWithMetadata",
     "InBatchSampler",
-    "CachedCrossBatchSampler",
-    "CachedUniformSampler",
     "PopularityBasedSampler",
     "FIFOQueue",
     "YoutubeDNNRetrievalModel",
