@@ -14,22 +14,16 @@
 # limitations under the License.
 #
 import logging
-from typing import Optional, Sequence, Union
+from typing import Dict, Optional, Sequence, Union
 
 import tensorflow as tf
 from tensorflow.python.ops import embedding_ops
 
-from merlin.models.tf.blocks.core.base import (
-    Block,
-    BlockType,
-    EmbeddingWithMetadata,
-    PredictionOutput,
-)
-from merlin.models.tf.blocks.core.combinators import ParallelBlock
-from merlin.models.tf.blocks.core.context import FeatureContext
-from merlin.models.tf.blocks.core.tabular import Filter, TabularAggregationType
-from merlin.models.tf.blocks.core.transformations import L2Norm
 from merlin.models.tf.blocks.sampling.base import ItemSampler
+from merlin.models.tf.core.base import Block, BlockType, EmbeddingWithMetadata, PredictionOutput
+from merlin.models.tf.core.combinators import ParallelBlock
+from merlin.models.tf.core.tabular import Filter, TabularAggregationType
+from merlin.models.tf.core.transformations import L2Norm
 from merlin.models.tf.models.base import ModelBlock
 from merlin.models.tf.typing import TabularData
 from merlin.models.tf.utils.tf_utils import (
@@ -76,7 +70,6 @@ class DualEncoderBlock(ParallelBlock):
         **kwargs,
     ):
         """Prepare the Query and Item towers of a Retrieval block
-
         Parameters
         ----------
         query_block : Block
@@ -268,7 +261,7 @@ class ItemRetrievalScorer(Block):
     def call_outputs(
         self,
         outputs: PredictionOutput,
-        feature_context: FeatureContext = None,
+        features: Dict[str, tf.Tensor] = None,
         training=True,
         testing=False,
         **kwargs,
@@ -295,7 +288,7 @@ class ItemRetrievalScorer(Block):
         if self.sampled_softmax_mode or isinstance(targets, tf.Tensor):
             positive_item_ids = targets
         else:
-            positive_item_ids = feature_context.features.values[self.item_id_feature_name]
+            positive_item_ids = features[self.item_id_feature_name]
 
         neg_items_ids = None
         if training or testing:
@@ -311,8 +304,7 @@ class ItemRetrievalScorer(Block):
 
             batch_items_embeddings = predictions[self.item_name]
             batch_items_metadata = {
-                feat_name: feature_context.features.values[feat_name]
-                for feat_name in self._required_features
+                feat_name: features[feat_name] for feat_name in self._required_features
             }
 
             positive_scores = tf.reduce_sum(
@@ -357,7 +349,7 @@ class ItemRetrievalScorer(Block):
                 if isinstance(targets, tf.Tensor):
                     positive_item_ids = targets
                 else:
-                    positive_item_ids = feature_context.features.values[self.item_id_feature_name]
+                    positive_item_ids = features[self.item_id_feature_name]
 
                 if len(neg_items_ids_list) == 1:
                     neg_items_ids = neg_items_ids_list[0]
@@ -443,8 +435,8 @@ class ItemRetrievalScorer(Block):
     def get_config(self):
         config = super().get_config()
         config = maybe_serialize_keras_objects(self, config, ["samplers"])
-        config["downscore_false_negatives"] = self.downscore_false_negatives
-        config["false_negatives_score"] = self.false_negatives_score
+        config["sampling_downscore_false_negatives"] = self.downscore_false_negatives
+        config["sampling_downscore_false_negatives_value"] = self.false_negatives_score
         config["item_id_feature_name"] = self.item_id_feature_name
 
         return config
