@@ -1,32 +1,35 @@
 from __future__ import annotations
 
 import inspect
+import sys
 from collections.abc import Sequence as SequenceCollection
 from typing import TYPE_CHECKING, Dict, List, Optional, Protocol, Union, runtime_checkable
 
+import six
 import tensorflow as tf
 from keras.utils.losses_utils import cast_losses_to_common_dtype
-from tensorflow.python.keras.engine import data_adapter
+from packaging import version
+from tensorflow.keras.utils import unpack_x_y_sample_weight
 
 import merlin.io
-from merlin.models.config.schema import FeatureCollection
-from merlin.models.tf.blocks.core.base import Block, ModelContext, PredictionOutput, is_input_block
-from merlin.models.tf.blocks.core.combinators import SequentialBlock
-from merlin.models.tf.blocks.core.context import FeatureContext
-from merlin.models.tf.blocks.core.transformations import AsDenseFeatures
+from merlin.models.tf.core.base import Block, ModelContext, PredictionOutput, is_input_block
+from merlin.models.tf.core.combinators import SequentialBlock
+from merlin.models.tf.core.prediction import Prediction, PredictionContext
+from merlin.models.tf.core.tabular import TabularBlock
+from merlin.models.tf.core.transformations import AsDenseFeatures, AsRaggedFeatures
 from merlin.models.tf.dataset import BatchedDataset
 from merlin.models.tf.inputs.base import InputBlock
 from merlin.models.tf.losses.base import loss_registry
-from merlin.models.tf.metrics.ranking import RankingMetric
+from merlin.models.tf.metrics.topk import filter_topk_metrics
 from merlin.models.tf.models.utils import parse_prediction_tasks
 from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
 from merlin.models.tf.utils.search_utils import find_all_instances_in_layers
-from merlin.models.tf.utils.tf_utils import call_layer
+from merlin.models.tf.utils.tf_utils import call_layer, maybe_deserialize_keras_objects
 from merlin.models.utils.dataset import unique_rows_by_features
 from merlin.schema import Schema, Tags
 
 if TYPE_CHECKING:
-    from merlin.models.tf.blocks.core.index import TopKIndexBlock
+    from merlin.models.tf.core.index import TopKIndexBlock
 
 
 class MetricsComputeCallback(tf.keras.callbacks.Callback):
@@ -66,7 +69,9 @@ class ModelBlock(Block, tf.keras.Model):
         self.block = block
 
     def call(self, inputs, **kwargs):
-        outputs = self.block(inputs, **kwargs)
+        if "features" not in kwargs:
+            kwargs["features"] = inputs
+        outputs = call_layer(self.block, inputs, **kwargs)
         return outputs
 
     def build(self, input_shapes):
@@ -162,50 +167,18 @@ class ModelBlock(Block, tf.keras.Model):
     def get_config(self):
         return {"block": tf.keras.utils.serialize_keras_object(self.block)}
 
+    def _set_save_spec(self, inputs, args=None, kwargs=None):
+        # We need to overwrite this in order to fix a Keras-bug in TF<2.9
+        super()._set_save_spec(inputs, args, kwargs)
 
-@tf.keras.utils.register_keras_serializable(package="merlin.models")
-class Model(tf.keras.Model):
-    def __init__(
-        self,
-        *blocks: Block,
-        context: Optional[ModelContext] = None,
-        **kwargs,
-    ):
-        super(Model, self).__init__(**kwargs)
-        context = context or ModelContext()
-        if len(blocks) == 1 and isinstance(blocks[0], SequentialBlock):
-            self.block = blocks[0]
-        else:
-            self.block = SequentialBlock(blocks, context=context, block_name="blocks")
-        if not getattr(self.block, "_context", None):
-            self.block._set_context(context)
-        self.context = context
-        self._is_fitting = False
+        if version.parse(tf.__version__) < version.parse("2.9.0"):
+            # Keras will interpret kwargs like `features` & `targets` as
+            # required args, which is wrong. This is a workaround.
+            _arg_spec = self._saved_model_arg_spec
+            self._saved_model_arg_spec = ([_arg_spec[0][0]], _arg_spec[1])
 
-        input_block_schemas = [
-            block.schema for block in self.block.submodules if getattr(block, "is_input", False)
-        ]
-        self.schema = sum(input_block_schemas, Schema())
-
-        self.as_dense = AsDenseFeatures()
-
-        # Initializing model control flags controlled by MetricsComputeCallback()
-        self._should_compute_train_metrics_for_batch = tf.Variable(
-            dtype=tf.bool,
-            name="should_compute_train_metrics_for_batch",
-            trainable=False,
-            synchronization=tf.VariableSynchronization.NONE,
-            initial_value=lambda: False,
-        )
-
-    def call(self, inputs, **kwargs):
-        if not kwargs.get("feature_context", None):
-            features = FeatureCollection(self.schema, self.as_dense(inputs))
-            kwargs["feature_context"] = FeatureContext(features)
-
-        outputs = call_layer(self.block, inputs, **kwargs)
-        return outputs
 
+class BaseModel(tf.keras.Model):
     def compile(
         self,
         optimizer="rmsprop",
@@ -306,6 +279,15 @@ class Model(tf.keras.Model):
             **kwargs: Arguments supported for backwards compatibility only.
         """
 
+        # Initializing model control flags controlled by MetricsComputeCallback()
+        self._should_compute_train_metrics_for_batch = tf.Variable(
+            dtype=tf.bool,
+            name="should_compute_train_metrics_for_batch",
+            trainable=False,
+            synchronization=tf.VariableSynchronization.NONE,
+            initial_value=lambda: False,
+        )
+
         self.output_names = [task.task_name for task in self.prediction_tasks]
 
         _metrics = {}
@@ -333,7 +315,7 @@ class Model(tf.keras.Model):
             if isinstance(_loss[key], str) and _loss[key] in loss_registry:
                 _loss[key] = loss_registry.parse(_loss[key])
 
-        super(Model, self).compile(
+        super(BaseModel, self).compile(
             optimizer=optimizer,
             loss=_loss,
             metrics=_metrics,
@@ -345,19 +327,11 @@ class Model(tf.keras.Model):
             **kwargs,
         )
 
-    @property
-    def first(self):
-        return self.block.layers[0]
-
-    @property
-    def last(self):
-        return self.block.layers[-1]
-
     @property
     def prediction_tasks(self) -> List[PredictionTask]:
         from merlin.models.tf.prediction_tasks.base import PredictionTask
 
-        results = find_all_instances_in_layers(self.block, PredictionTask)
+        results = find_all_instances_in_layers(self, PredictionTask)
 
         return results
 
@@ -376,58 +350,19 @@ class Model(tf.keras.Model):
 
         return outputs
 
-    @classmethod
-    def from_block(
-        cls,
-        block: Block,
-        schema: Schema,
-        input_block: Optional[Block] = None,
-        prediction_tasks: Optional[
-            Union["PredictionTask", List["PredictionTask"], "ParallelPredictionBlock"]
-        ] = None,
-        aggregation="concat",
-        **kwargs,
-    ) -> "Model":
-        """Create a model from a `block`
-
-        Parameters
-        ----------
-        block: Block
-            The block to wrap in-between an InputBlock and prediction task(s)
-        schema: Schema
-            Schema to use for the model.
-        input_block: Optional[Block]
-            Block to use as input.
-        prediction_tasks: Optional[
-            Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
-        ]
-            Prediction tasks to use.
-        """
-        if is_input_block(block.first):
-            if input_block is not None:
-                raise ValueError("The block already includes an InputBlock")
-            input_block = block.first
-
-        _input_block: Block = input_block or InputBlock(schema, aggregation=aggregation, **kwargs)
-
-        prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
-
-        return cls(_input_block, block, prediction_tasks)
-
-    def prediction_output(
+    def call_train_test(
         self, x, y=None, training=False, testing=False, **kwargs
     ) -> PredictionOutput:
-        features = FeatureCollection(self.schema, self.as_dense(x))
-        feature_context = FeatureContext(features)
-
         forward = self(
             x,
             targets=y,
             training=training,
             testing=testing,
-            feature_context=feature_context,
             **kwargs,
         )
+        if not self.prediction_tasks:
+            return PredictionOutput(forward, y)
+
         predictions, targets, output = {}, {}, None
         for task in self.prediction_tasks:
             task_x = forward
@@ -456,8 +391,8 @@ class Model(tf.keras.Model):
         """Custom train step using the `compute_loss` method."""
 
         with tf.GradientTape() as tape:
-            x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
-            outputs = self.prediction_output(x, y, training=True)
+            x, y, sample_weight = unpack_x_y_sample_weight(data)
+            outputs = self.call_train_test(x, y, training=True)
             loss = self.compute_loss(x, outputs.targets, outputs.predictions, sample_weight)
 
         self._validate_target_and_loss(outputs.targets, loss)
@@ -465,9 +400,7 @@ class Model(tf.keras.Model):
         # Run backwards pass.
         self.optimizer.minimize(loss, self.trainable_variables, tape=tape)
 
-        metrics = self.compute_metrics(
-            x, outputs.targets, outputs.predictions, sample_weight, training=True
-        )
+        metrics = self.compute_metrics(outputs, sample_weight, training=True)
         # Adding regularization loss to metrics
         metrics["regularization_loss"] = tf.reduce_sum(cast_losses_to_common_dtype(self.losses))
 
@@ -476,8 +409,8 @@ class Model(tf.keras.Model):
     def test_step(self, data):
         """Custom test step using the `compute_loss` method."""
 
-        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
-        outputs = self.prediction_output(x, y, testing=True)
+        x, y, sample_weight = unpack_x_y_sample_weight(data)
+        outputs = self.call_train_test(x, y, testing=True)
 
         if getattr(self, "pre_eval_topk", None) is not None:
             # During eval, the retrieval-task only returns positive scores
@@ -485,9 +418,7 @@ class Model(tf.keras.Model):
             outputs = self.pre_eval_topk.call_outputs(outputs)
 
         self.compute_loss(x, outputs.targets, outputs.predictions, sample_weight)
-        metrics = self.compute_metrics(
-            x, outputs.targets, outputs.predictions, sample_weight, training=False
-        )
+        metrics = self.compute_metrics(outputs, sample_weight, training=False)
         # Adding regularization loss to metrics
         metrics["regularization_loss"] = tf.reduce_sum(cast_losses_to_common_dtype(self.losses))
 
@@ -496,25 +427,18 @@ class Model(tf.keras.Model):
     @tf.function
     def compute_metrics(
         self,
-        x: tf.Tensor,
-        y: tf.Tensor,
-        y_pred: tf.Tensor,
+        prediction_outputs: PredictionOutput,
         sample_weight: tf.Tensor,
         training: bool,
     ) -> Dict[str, tf.Tensor]:
-        """Overrides default Keras Model.compute_metrics()
-           to compute metrics each N steps (and not for all steps)
-           during training
+        """Overrides Model.compute_metrics() for some custom behaviour
+           like compute metrics each N steps during training
+           and allowing to feed additional information required by specific metrics
 
         Parameters
         ----------
-        x : tf.Tensor
-            Input features. Not used, and kept only to align with the signature
-            of super Model.compute_metrics()
-        y : tf.Tensor
-            Target values
-        y_pred : tf.Tensor
-            Predictions values
+        prediction_outputs : PredictionOutput
+            Contains properties with targets and predictions
         sample_weight : tf.Tensor
             Sample weights for weighted metrics
         training : bool
@@ -526,10 +450,26 @@ class Model(tf.keras.Model):
         Dict[str, tf.Tensor]
             Dict with the metrics values
         """
-        del x
-        # During training updates metric states each N steps
-        if self._should_compute_train_metrics_for_batch or not training:
-            self.compiled_metrics.update_state(y, y_pred, sample_weight)
+
+        should_compute_metrics = self._should_compute_train_metrics_for_batch or not training
+        if should_compute_metrics:
+
+            # This ensures that compiled metrics are built
+            # to make self.compiled_metrics.metrics available
+            if not self.compiled_metrics.built:
+                self.compiled_metrics.build(
+                    prediction_outputs.predictions, prediction_outputs.targets
+                )
+
+            # Providing label_relevant_counts for TopkMetrics, as metric.update_state()
+            # should have standard signature for better compatibility with Keras methods
+            # like self.compiled_metrics.update_state()
+            for topk_metric in filter_topk_metrics(self.compiled_metrics.metrics):
+                topk_metric.label_relevant_counts = prediction_outputs.label_relevant_counts
+
+            self.compiled_metrics.update_state(
+                prediction_outputs.targets, prediction_outputs.predictions, sample_weight
+            )
         # Returns the current value of metrics
         metrics = self.metrics_results()
         return metrics
@@ -577,6 +517,11 @@ class Model(tf.keras.Model):
         **kwargs,
     ):
         x = _maybe_convert_merlin_dataset(x, batch_size, **kwargs)
+
+        # Bind schema from dataset to model in case we can't infer it from the inputs
+        if isinstance(x, BatchedDataset):
+            self.schema = x.schema
+
         validation_data = _maybe_convert_merlin_dataset(
             validation_data, batch_size, shuffle=shuffle, **kwargs
         )
@@ -652,7 +597,7 @@ class Model(tf.keras.Model):
     ):
         x = _maybe_convert_merlin_dataset(x, batch_size, shuffle=False, **kwargs)
 
-        return super(Model, self).predict(
+        return super(BaseModel, self).predict(
             x,
             batch_size=batch_size,
             verbose=verbose,
@@ -694,14 +639,214 @@ class Model(tf.keras.Model):
 
         return merlin.io.Dataset(predictions)
 
+
+@tf.keras.utils.register_keras_serializable(package="merlin.models")
+class Model(BaseModel):
+    def __init__(
+        self,
+        *blocks: Block,
+        context: Optional[ModelContext] = None,
+        pre: Optional[tf.keras.layers.Layer] = None,
+        post: Optional[tf.keras.layers.Layer] = None,
+        **kwargs,
+    ):
+        super(Model, self).__init__(**kwargs)
+        context = context or ModelContext()
+        if len(blocks) == 1 and isinstance(blocks[0], SequentialBlock):
+            blocks = blocks[0].layers
+
+        self.blocks = blocks
+        for block in self.submodules:
+            if hasattr(block, "_set_context"):
+                block._set_context(context)
+
+        self.pre = pre
+        self.post = post
+        self.context = context
+        self._is_fitting = False
+
+        input_block_schemas = [
+            block.schema for block in self.submodules if getattr(block, "is_input", False)
+        ]
+        self.schema = sum(input_block_schemas, Schema())
+
+    def _maybe_build(self, inputs):
+        if isinstance(inputs, dict):
+            _ragged_inputs = AsRaggedFeatures()(inputs)
+            feature_shapes = {k: v.shape for k, v in _ragged_inputs.items()}
+            feature_dtypes = {k: v.dtype for k, v in _ragged_inputs.items()}
+
+            for block in self.blocks:
+                block._feature_shapes = feature_shapes
+                block._feature_dtypes = feature_dtypes
+                for child in block.submodules:
+                    child._feature_shapes = feature_shapes
+                    child._feature_dtypes = feature_dtypes
+        super()._maybe_build(inputs)
+
+    def build(self, input_shape=None):
+        """Builds the model
+
+        Parameters
+        ----------
+        input_shape : tf.TensorShape, optional
+            The input shape, by default None
+        """
+        last_layer = None
+
+        if self.pre is not None:
+            self.pre.build(input_shape)
+            input_shape = self.pre.compute_output_shape(input_shape)
+
+        for layer in self.blocks:
+            try:
+                layer.build(input_shape)
+            except TypeError:
+                t, v, tb = sys.exc_info()
+                if isinstance(input_shape, dict) and isinstance(last_layer, TabularBlock):
+                    v = TypeError(
+                        f"Couldn't build {layer}, "
+                        f"did you forget to add aggregation to {last_layer}?"
+                    )
+                six.reraise(t, v, tb)
+            input_shape = layer.compute_output_shape(input_shape)
+            last_layer = layer
+
+        if self.post is not None:
+            self.post.build(input_shape)
+
+        self.built = True
+
+    def call(self, inputs, targets=None, training=False, testing=False, output_context=False):
+        context = self._create_context(
+            AsDenseFeatures()(inputs),  # TODO: Change this to ragged
+            targets=targets,
+            training=training,
+            testing=testing,
+        )
+
+        outputs = inputs
+        if self.pre:
+            outputs, context = self._call_child(self.pre, outputs, context)
+
+        for block in self.blocks:
+            outputs, context = self._call_child(block, outputs, context)
+
+        if self.post:
+            outputs, context = self._call_child(self.post, outputs, context)
+
+        if output_context:
+            return outputs, context
+
+        return outputs
+
+    def _create_context(
+        self, inputs, targets=None, training=False, testing=False
+    ) -> PredictionContext:
+        context = PredictionContext(inputs, targets=targets, training=training, testing=testing)
+
+        return context
+
+    def _call_child(
+        self,
+        child: tf.keras.layers.Layer,
+        inputs,
+        context: PredictionContext,
+    ):
+        call_kwargs = context.to_call_dict()
+
+        # Prevent features to be part of signature of model-blocks
+        if any(isinstance(sub, ModelBlock) for sub in child.submodules):
+            del call_kwargs["features"]
+
+        outputs = call_layer(child, inputs, **call_kwargs)
+        if isinstance(outputs, Prediction):
+            targets = outputs.targets if outputs.targets is not None else context.targets
+            features = outputs.features if outputs.features is not None else context.features
+            outputs = outputs[0]
+            context = context.with_updates(targets=targets, features=features)
+
+        return outputs, context
+
+    @property
+    def first(self):
+        return self.blocks[0]
+
+    @property
+    def last(self):
+        return self.blocks[-1]
+
+    @classmethod
+    def from_block(
+        cls,
+        block: Block,
+        schema: Schema,
+        input_block: Optional[Block] = None,
+        prediction_tasks: Optional[
+            Union["PredictionTask", List["PredictionTask"], "ParallelPredictionBlock"]
+        ] = None,
+        aggregation="concat",
+        **kwargs,
+    ) -> "Model":
+        """Create a model from a `block`
+
+        Parameters
+        ----------
+        block: Block
+            The block to wrap in-between an InputBlock and prediction task(s)
+        schema: Schema
+            Schema to use for the model.
+        input_block: Optional[Block]
+            Block to use as input.
+        prediction_tasks: Optional[
+            Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
+        ]
+            Prediction tasks to use.
+        """
+        if isinstance(block, SequentialBlock) and is_input_block(block.first):
+            if input_block is not None:
+                raise ValueError("The block already includes an InputBlock")
+            input_block = block.first
+
+        _input_block: Block = input_block or InputBlock(schema, aggregation=aggregation, **kwargs)
+
+        prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
+
+        return cls(_input_block, block, prediction_tasks)
+
     @classmethod
     def from_config(cls, config, custom_objects=None):
-        block = tf.keras.utils.deserialize_keras_object(config.pop("block"))
+        pre = config.pop("pre", None)
+        post = config.pop("post", None)
+        layers = [
+            tf.keras.layers.deserialize(conf, custom_objects=custom_objects)
+            for conf in config.values()
+        ]
 
-        return cls(block, **config)
+        if pre is not None:
+            pre = tf.keras.layers.deserialize(pre, custom_objects=custom_objects)
+
+        if post is not None:
+            post = tf.keras.layers.deserialize(post, custom_objects=custom_objects)
+
+        return cls(*layers, pre=pre, post=post)
 
     def get_config(self):
-        return {"block": tf.keras.utils.serialize_keras_object(self.block)}
+        config = maybe_deserialize_keras_objects({}, ["pre", "post"])
+        for i, layer in enumerate(self.blocks):
+            config[i] = tf.keras.utils.serialize_keras_object(layer)
+
+        return config
+
+    def _set_save_spec(self, inputs, args=None, kwargs=None):
+        # We need to overwrite this in order to fix a Keras-bug in TF<2.9
+        super()._set_save_spec(inputs, args, kwargs)
+
+        if version.parse(tf.__version__) < version.parse("2.9.0"):
+            # Keras will interpret kwargs like `features` & `targets` as
+            # required args, which is wrong. This is a workaround.
+            _arg_spec = self._saved_model_arg_spec
+            self._saved_model_arg_spec = ([_arg_spec[0][0]], _arg_spec[1])
 
 
 @runtime_checkable
@@ -733,11 +878,10 @@ class RetrievalModel(Model):
         return_dict=False,
         **kwargs,
     ):
-        self.has_ranking_metric = any(isinstance(m, RankingMetric) for m in self.metrics)
         self.has_item_corpus = False
 
         if item_corpus:
-            from merlin.models.tf.blocks.core.index import TopKIndexBlock
+            from merlin.models.tf.core.index import TopKIndexBlock
 
             self.has_item_corpus = True
 
@@ -748,13 +892,18 @@ class RetrievalModel(Model):
                 item_block = self.retrieval_block.item_block()
 
                 if not getattr(self, "pre_eval_topk", None):
-                    ranking_metrics = list(
-                        [metric for metric in self.metrics if isinstance(metric, RankingMetric)]
-                    )
+                    topk_metrics = filter_topk_metrics(self.metrics)
+                    if len(topk_metrics) == 0:
+                        # TODO: Decouple the evaluation of RetrievalModel from the need of using
+                        # at least one TopkMetric (how to infer the k for TopKIndexBlock?)
+                        raise ValueError(
+                            "RetrievalModel evaluation requires at least "
+                            "one TopkMetric (e.g., RecallAt(5), NDCGAt(10))."
+                        )
                     self.pre_eval_topk = TopKIndexBlock.from_block(
                         item_block,
                         data=item_corpus,
-                        k=tf.reduce_max([metric.k for metric in ranking_metrics]),
+                        k=tf.reduce_max([metric.k for metric in topk_metrics]),
                         context=self.context,
                         **kwargs,
                     )
@@ -789,7 +938,7 @@ class RetrievalModel(Model):
 
     @property
     def retrieval_block(self) -> RetrievalBlock:
-        return next(b for b in self.block if isinstance(b, RetrievalBlock))
+        return next(b for b in self.blocks if isinstance(b, RetrievalBlock))
 
     def query_embeddings(
         self,
@@ -892,11 +1041,9 @@ class RetrievalModel(Model):
 
         if isinstance(item_corpus, merlin.io.Dataset):
             if not k:
-                ranking_metrics = list(
-                    [metric for metric in self.metrics if isinstance(metric, RankingMetric)]
-                )
-                if ranking_metrics:
-                    k = tf.reduce_max([metric.k for metric in ranking_metrics])
+                topk_metrics = filter_topk_metrics(self.metrics)
+                if topk_metrics:
+                    k = tf.reduce_max([metric.k for metric in topk_metrics])
                 else:
                     raise ValueError("You must specify a k for the Top-k Recommender.")
 
