@@ -13,16 +13,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-from typing import Any, Sequence, Union
+from typing import Any, Sequence, Tuple, Union
 
 import numpy as np
 import tensorflow as tf
+from keras.utils.tf_inspect import getfullargspec
 from packaging import version
 
 from merlin.core.dispatch import DataFrameType
 from merlin.io import Dataset
 from merlin.models.tf.typing import TabularData
-from merlin.models.tf.utils import tf_utils
 from merlin.models.utils.misc_utils import filter_kwargs
 
 if version.parse(tf.__version__) < version.parse("2.3.0"):
@@ -231,6 +231,10 @@ def df_to_tensor(gdf, dtype=None):
         # matrix which means we had to transpose
         # for the bug above, so untranspose
         x = tf.transpose(x)
+
+    if dtype:
+        return tf.cast(x, dtype)
+
     return x
 
 
@@ -292,6 +296,7 @@ def get_candidate_probs(
     return candidate_probs
 
 
+@tf.keras.utils.register_keras_serializable(package="merlin.models")
 class TensorInitializer(tf.keras.initializers.Initializer):
     """Initializer that returns a tensor (e.g. pre-trained
     embeddings) set in the constructor
@@ -325,27 +330,45 @@ class TensorInitializer(tf.keras.initializers.Initializer):
         return weights
 
     @classmethod
-    def from_dataset(cls, data: Union[Dataset, DataFrameType], **kwargs):
+    def from_dataset(cls, data: Union[Dataset, DataFrameType], **kwargs) -> "TensorInitializer":
         if hasattr(data, "to_ddf"):
             data = data.to_ddf().compute()
-        embeddings = tf_utils.df_to_tensor(data)
+        embeddings = df_to_tensor(data)
 
         return cls(weights=embeddings, **kwargs)
 
     def get_config(self):  # To support serialization
-        return {"weights": self._weights}
+        return {"weights": self._weights.numpy()}
 
 
 def call_layer(layer: tf.keras.layers.Layer, inputs, *args, **kwargs):
     """Calls a layer with the given inputs and filters kwargs. Returns the output"""
 
     has_custom_call = getattr(layer, "_has_custom__call__", False)
+    _k = dict(cascade_kwargs_if_possible=True, argspec_fn=getfullargspec)
 
-    filtered_kwargs = filter_kwargs(kwargs, layer, cascade_kwargs_if_possible=True)
+    filtered_kwargs = filter_kwargs(kwargs, layer, **_k)
 
     if not has_custom_call:
-        filtered_kwargs = filter_kwargs(
-            filtered_kwargs, layer.call, cascade_kwargs_if_possible=True
-        )
+        if isinstance(layer, tf.keras.layers.Lambda):
+            filtered_kwargs = filter_kwargs(kwargs, layer.function, **_k)
+        else:
+            # We need to check the call method on the type since when the model gets saved
+            # we can't infer the kwargs from using `layer.call` directly
+            call_fn = type(layer).call
+
+            filtered_kwargs = filter_kwargs(filtered_kwargs, call_fn, **_k)
 
     return layer(inputs, *args, **filtered_kwargs)
+
+
+def list_col_to_ragged(col: Tuple[tf.Tensor, tf.Tensor]):
+    values = col[0][:, 0]
+    row_lengths = col[1][:, 0]
+
+    if values.dtype.is_floating:
+        values = tf.cast(values, tf.int32)
+    if row_lengths.dtype.is_floating:
+        row_lengths = tf.cast(row_lengths, tf.int32)
+
+    return tf.RaggedTensor.from_row_lengths(values, row_lengths)
